#include <cassert>
#include <cstring>
#include <iostream>
#include <rapidjson/document.h>
#include <rapidjson/error/en.h>

void json_has_parse_error()
{
    char const *const json = R"({"bad":{})";

    rapidjson::Document d;
    d.Parse(json);

    if (d.HasParseError())
    {
        std::cout << "Parse Error at offset " << d.GetErrorOffset()
                  << ". "
                  << rapidjson::GetParseError_En(d.GetParseError())
                  << '\n';
    }
}

void access_array_with_elements(rapidjson::Value const &j)
{
    rapidjson::Value::ConstMemberIterator const awe_itr =
        j.FindMember("array_with_elements");

    assert(awe_itr != j.MemberEnd());
    assert(awe_itr->value.IsArray());
    assert(awe_itr->value.Empty() == false);
    assert(awe_itr->value.Size() == 3);

    rapidjson::Value::ConstArray const a = awe_itr->value.GetArray();

    // By index
    for (rapidjson::SizeType i = 0; i < a.Size(); ++i)
    {
        assert(a[i].IsString());
    }

    // By iterator
    for (rapidjson::Value::ConstValueIterator element_itr = a.Begin();
         element_itr != a.End();
         ++element_itr)
    {
        assert(element_itr->IsString());
    }
}

void access_string(rapidjson::Value const &j)
{
    rapidjson::Value::ConstMemberIterator const s_itr =
        j.FindMember("string");

    assert(s_itr != j.MemberEnd());
    assert(s_itr->value.IsString());

    assert(std::strcmp(s_itr->value.GetString(), "hello world") == 0);
}

void access_int(rapidjson::Value const &j)
{
    rapidjson::Value::ConstMemberIterator const i_itr =
        j.FindMember("int");

    assert(i_itr != j.MemberEnd());
    assert(i_itr->value.IsNumber());
    assert(i_itr->value.IsInt());
    assert(i_itr->value.IsInt64());
    assert(i_itr->value == 42);
}

void access_float(rapidjson::Value const &j)
{
    rapidjson::Value::ConstMemberIterator const f_itr =
        j.FindMember("float");

    assert(f_itr != j.MemberEnd());
    assert(f_itr->value.IsNumber());
    assert(f_itr->value.IsDouble());
    assert(f_itr->value == 42.42);
}

void access_boolean(rapidjson::Value const &j)
{
    rapidjson::Value::ConstMemberIterator const f_itr =
        j.FindMember("boolean");

    assert(f_itr != j.MemberEnd());
    assert(f_itr->value.IsBool());
    assert(f_itr->value.IsTrue() == false);
    assert(f_itr->value.IsFalse());
    assert(f_itr->value == false);
}

void incorrectly_access_string_as_object(rapidjson::Value const &d)
{
    const char *const node_name = "string";

    assert(d[node_name].IsString());

#if 0// Invalid - assertion will trigger
    d[node_name].GetObject();
#endif
}

void access_object_with_elements(rapidjson::Value const &j)
{
    const char *const node_name = "object_with_elements";
    assert(j.HasMember(node_name));

    assert(j[node_name].IsObject());
    assert(j[node_name].ObjectEmpty() == false);
    assert(j[node_name].MemberCount() == 2);

    assert(j[node_name]["a"].IsNumber());
    assert(j[node_name]["a"] == 1);

    assert(j[node_name]["b"].IsNumber());
    assert(j[node_name]["b"] == 2);
}

rapidjson::Document parse_test_document()
{
    char const *const json = R"(
    {
        "array_with_elements": ["one", "two", "three"],
        "object_with_elements" : {
            "a" : 1,
            "b" : 2
        },
        "string": "hello world",
        "int": 42,
        "float": 42.42,
        "boolean": false
    }
    )";

    rapidjson::Document d;
    d.Parse(json);

    assert(d.HasParseError() == false);
    assert(d.GetParseError() == rapidjson::kParseErrorNone);

    return d;
}

void get_doc_as_object(rapidjson::Document const &d)
{
    assert(d.IsObject());
    assert(d.GetType() == rapidjson::kObjectType);
    assert(d.MemberCount() == 6);

    rapidjson::Value const &v(d);
    assert(v.GetType() == rapidjson::kObjectType);
    assert(v.IsObject());
    assert(v.MemberCount() == 6);

    rapidjson::Value::ConstObject const o = d.GetObject();
    assert(o.MemberCount() == 6);
}

void access_object_using_iterator(rapidjson::Value const &d)
{
    rapidjson::Value::ConstMemberIterator const owe_itr =
        d.FindMember("object_with_elements");

    assert(owe_itr != d.MemberEnd() && owe_itr->value.IsObject());

    rapidjson::Value const &owe = owe_itr->value;
    assert(owe.MemberCount() == 2);

    rapidjson::Value::ConstMemberIterator member_itr;

    member_itr = owe.FindMember("a");
    assert(member_itr != owe.MemberEnd());
    assert(member_itr->value.IsNumber());
    assert(member_itr->value == 1);

    member_itr = owe.FindMember("b");
    assert(member_itr != owe.MemberEnd());
    assert(member_itr->value.IsNumber());
    assert(member_itr->value == 2);
}

int main()
{
    json_has_parse_error();

    rapidjson::Document const d = parse_test_document();

    get_doc_as_object(d);

    access_array_with_elements(d);

    access_object_with_elements(d);

    access_object_using_iterator(d);

    access_string(d);

    access_int(d);

    access_float(d);

    access_boolean(d);

    incorrectly_access_string_as_object(d);

    return 0;
}
